@@ -2,19 +2,13 @@
 /* vim: set filetype=c tabstop=8 shiftwidth=4 expandtab: */
 
 /*
-    internal.h -- Structures and functions that are not meant for the end user
-*/
-/*
-    Copyright (c) 2001, Juan Jose Garcia Ripoll.
-
-    ECL is free software; you can redistribute it and/or
-    modify it under the terms of the GNU Library General Public
-    License as published by the Free Software Foundation; either
-    version 2 of the License, or (at your option) any later version.
-
-    See file '../Copyright' for full details.
-*/
+ * Copyright (c) 2001, Juan Jose Garcia Ripoll.
+ *
+ * See file 'LICENSE' for the copyright details.
+ *
+ */
 
+/* internal.h -- Structures and functions that are not meant for the end user */
 #ifndef ECL_INTERNAL_H
 #define ECL_INTERNAL_H
 
@@ -48,6 +42,8 @@ extern void init_unixtime(void);
 extern void init_compiler(void);
 #ifdef ECL_THREADS
 extern void init_threads(cl_env_ptr);
+#else
+#define init_threads(env) cl_env_p = env
 #endif
 extern void ecl_init_env(cl_env_ptr);
 extern void init_lib_LSP(cl_object);
@@ -63,6 +59,7 @@ extern void _ecl_dealloc_env(cl_env_ptr);
 extern void _ecl_set_max_heap_size(size_t new_size);
 extern cl_object ecl_alloc_bytecodes(cl_index data_size, cl_index code_size);
 extern cl_index ecl_object_byte_size(cl_type t);
+extern cl_index ecl_next_stamp();
 
 /* array.d */
 
@@ -72,6 +69,12 @@ static const cl_index ecl_aet_size[] = {
   sizeof(cl_object),          /* ecl_aet_object */
   sizeof(float),              /* ecl_aet_sf */
   sizeof(double),             /* ecl_aet_df */
+  sizeof(long double),        /* ecl_aet_lf */
+#ifdef ECL_COMPLEX_FLOAT
+  sizeof(_Complex float),        /* ecl_aet_csf */
+  sizeof(_Complex double),       /* ecl_aet_cdf */
+  sizeof(_Complex long double),  /* ecl_aet_clf */
+#endif
   0,                          /* ecl_aet_bit: cannot be handled with this code */
   sizeof(cl_fixnum),          /* ecl_aet_fix */
   sizeof(cl_index),           /* ecl_aet_index */
@@ -106,12 +109,19 @@ struct cl_compiler_env {
         cl_fixnum lexical_level;        /* =0 if toplevel form */
         cl_object constants;            /* Constants for this form */
         cl_object load_time_forms;      /* Constants that have to be rebuilt */
+        cl_object ltf_being_created;    /* Load time objects being compiled */
+        cl_object ltf_defer_init_until; /* Defer evaluation of current
+                                         * load time init form until
+                                         * this object has been created */
+        cl_object ltf_locations;        /* Locations of constants externalized
+                                         * with make-load-form */
         cl_object lex_env;              /* Lexical env. for eval-when */
         cl_object code_walker;          /* Value of SI:*CODE-WALKER* */
         cl_index env_depth;
         cl_index env_size;
         int mode;
         bool stepping;
+        bool function_boundary_crossed;
 };
 
 typedef struct cl_compiler_env *cl_compiler_env_ptr;
@@ -160,17 +170,6 @@ extern cl_object si_constant_form_value _ECL_ARGS((cl_narg narg, cl_object form,
         struct ecl_stack_frame frame;\
         cl_object name = ecl_stack_frame_open(env, (cl_object)&frame, 0);
 
-#ifdef ECL_USE_VARARG_AS_POINTER
-#define ECL_STACK_FRAME_FROM_VA_LIST(e,f,va) do {                  \
-                const cl_object __frame = (f);                     \
-                __frame->frame.t = t_frame;                        \
-                __frame->frame.stack = 0;                          \
-                __frame->frame.env = (e);                          \
-                __frame->frame.size = va[0].narg;                  \
-                __frame->frame.base = va[0].sp? va[0].sp :         \
-                        (cl_object*)va[0].args;                    \
-        } while(0)
-#else
 #define ECL_STACK_FRAME_FROM_VA_LIST(e,f,va) do {                       \
                 const cl_object __frame = (f);                          \
                 cl_index i, __nargs = va[0].narg;                       \
@@ -179,35 +178,15 @@ extern cl_object si_constant_form_value _ECL_ARGS((cl_narg narg, cl_object form,
                         __frame->frame.base[i] = ecl_va_arg(va);         \
                 }                                                       \
         } while (0)
-#endif
 
-#ifdef ECL_USE_VARARG_AS_POINTER
 #define ECL_STACK_FRAME_VARARGS_BEGIN(narg,lastarg,frame)               \
-        struct ecl_frame __ecl_frame;                                   \
-        const cl_object frame = (cl_object)&__ecl_frame;                \
-        const cl_env_ptr env = ecl_process_env();                       \
-        frame->frame.t = t_frame;                                       \
-        frame->frame.stack = 0;                                         \
-        frame->frame.env = env;                                         \
-        frame->frame.size = narg;                                       \
-        if (narg < ECL_C_ARGUMENTS_LIMIT) {                             \
-                va_list args;                                           \
-                va_start(args, lastarg);                                \
-                frame->frame.base = (cl_object*)args;                   \
-        } else {                                                        \
-                frame->frame.base = env->stack_top - narg;              \
-        }
-#define ECL_STACK_FRAME_VARARGS_END(frame)      \
-        /* No stack consumed, no need to close frame */
-#else
-#define ECL_STACK_FRAME_VARARGS_BEGIN(narg,lastarg,frame)               \
-        struct ecl_frame __ecl_frame;                                   \
+        struct ecl_stack_frame __ecl_frame;                             \
         const cl_object frame = (cl_object)&__ecl_frame;                \
         const cl_env_ptr env = ecl_process_env();                       \
         frame->frame.t = t_frame;                                       \
         frame->frame.env = env;                                         \
         frame->frame.size = narg;                                       \
-        if (narg < ECL_C_ARGUMENTS_LIMIT) {                                 \
+        if (narg <= ECL_C_ARGUMENTS_LIMIT) {                            \
                 cl_object *p = frame->frame.base = env->values;         \
                 va_list args;                                           \
                 va_start(args, lastarg);                                \
@@ -215,6 +194,7 @@ extern cl_object si_constant_form_value _ECL_ARGS((cl_narg narg, cl_object form,
                         *p = va_arg(args, cl_object);                   \
                         ++p;                                            \
                 }                                                       \
+                va_end(args);                                           \
                 frame->frame.stack = (cl_object*)0x1;                   \
         } else {                                                        \
                 frame->frame.base = env->stack_top - narg;              \
@@ -222,10 +202,10 @@ extern cl_object si_constant_form_value _ECL_ARGS((cl_narg narg, cl_object form,
         }
 #define ECL_STACK_FRAME_VARARGS_END(frame)      \
         /* No stack consumed, no need to close frame */
-#endif
 
 extern cl_object _ecl_bytecodes_dispatch_vararg(cl_narg narg, ...);
 extern cl_object _ecl_bclosure_dispatch_vararg(cl_narg narg, ...);
+extern cl_object ecl_close_around(cl_object fun, cl_object env);
 
 /* ffi/backtrace.d */
 
@@ -233,32 +213,21 @@ extern void _ecl_dump_c_backtrace();
 
 /* ffi.d */
 
-struct ecl_fficall {
-        char *buffer_sp;
-        size_t buffer_size;
-        union ecl_ffi_values output;
-        enum ecl_ffi_calling_convention cc;
-        struct ecl_fficall_reg *registers;
-        char buffer[ECL_FFICALL_LIMIT];
-        cl_object cstring;
-};
-
 extern enum ecl_ffi_tag ecl_foreign_type_code(cl_object type);
 
 /* file.d */
 
-/*
- * POSIX specifies that the "b" flag is ignored. This is good, because
- * under MSDOS and Apple's OS we need to open text files in binary mode,
- * so that we get both the carriage return and the linefeed characters.
- * Otherwise, it would be complicated to implement file-position and
- * seek operations.
- */
-#define OPEN_R  "rb"
-#define OPEN_W  "wb"
-#define OPEN_RW "r+b"
-#define OPEN_A  "ab"
-#define OPEN_RA "a+b"
+/* Windows does not have this flag (POSIX thing) */
+#ifndef O_CLOEXEC
+#define O_CLOEXEC 0
+#endif
+#ifndef O_NONBLOCK
+#define O_NONBLOCK 0
+#endif
+/* Windows needs to be told explicitely to open files in binary mode */
+#ifndef O_BINARY
+#define O_BINARY 0
+#endif
 
 #define ECL_FILE_STREAM_P(strm) \
         (ECL_ANSI_STREAM_P(strm) && (strm)->stream.mode < ecl_smm_synonym)
@@ -307,6 +276,9 @@ extern cl_object si_formatter_aux _ECL_ARGS((cl_narg narg, cl_object strm, cl_ob
 
 /* hash.d */
 extern cl_object ecl_extend_hashtable(cl_object hashtable);
+#ifdef ECL_EXTERNALIZABLE
+extern void ecl_reconstruct_serialized_hashtable(cl_object h);
+#endif
 
 /* gfun.d, kernel.lsp */
 
@@ -315,8 +287,8 @@ extern cl_object ecl_extend_hashtable(cl_object hashtable);
 #define GFUN_COMB(x) ((x)->instance.slots[2])
 
 extern cl_object FEnot_funcallable_vararg(cl_narg narg, ...);
-extern cl_object ecl_slot_reader_dispatch(cl_narg narg, cl_object instance);
-extern cl_object ecl_slot_writer_dispatch(cl_narg narg, cl_object value, cl_object instance);
+extern cl_object ecl_slot_reader_dispatch(cl_narg narg, ... /* cl_object instance */);
+extern cl_object ecl_slot_writer_dispatch(cl_narg narg, ... /* cl_object value, cl_object instance */);
 
 /* load.d */
 
@@ -327,18 +299,23 @@ extern cl_object _ecl_library_default_entry(void);
 
 extern cl_object _ecl_double_to_integer(double d);
 extern cl_object _ecl_float_to_integer(float d);
-#ifdef ECL_LONG_FLOAT
 extern cl_object _ecl_long_double_to_integer(long double d);
-#endif
 
 /* main.d */
 
 extern cl_fixnum ecl_option_values[ECL_OPT_LIMIT+1];
 
+extern void ecl_init_bignum_registers(cl_env_ptr env);
+extern void ecl_clear_bignum_registers(cl_env_ptr env);
+
+/* threads/mutex.d */
+
+extern cl_object si_mutex_timeout();
+
 /* print.d */
 
 extern cl_object _ecl_stream_or_default_output(cl_object stream);
-extern void _ecl_write_addr(cl_object x, cl_object stream);
+extern void _ecl_write_addr(void *x, cl_object stream);
 extern void _ecl_write_array(cl_object o, cl_object stream);
 extern void _ecl_write_vector(cl_object o, cl_object stream);
 extern void _ecl_write_bitvector(cl_object o, cl_object stream);
@@ -360,71 +337,11 @@ extern void _ecl_string_push_c_string(cl_object s, const char *c);
 
 extern void cl_write_object(cl_object x, cl_object stream);
 
-/* global locks */
+/* threads/rwlock.d */
 
 #ifdef ECL_THREADS
-# define ECL_WITH_GLOBAL_LOCK_BEGIN(the_env)                    \
-        ECL_WITH_LOCK_BEGIN(the_env, cl_core.global_lock)
-# define ECL_WITH_GLOBAL_LOCK_END               \
-        ECL_WITH_LOCK_END
-# define ECL_WITH_LOCK_BEGIN(the_env,lock) {            \
-        const cl_env_ptr __ecl_the_env = the_env;       \
-        const cl_object __ecl_the_lock = lock;          \
-        ecl_disable_interrupts_env(the_env);            \
-        mp_get_lock_wait(__ecl_the_lock);               \
-        ECL_UNWIND_PROTECT_BEGIN(__ecl_the_env);                \
-        ecl_enable_interrupts_env(__ecl_the_env);
-# define ECL_WITH_LOCK_END                                    \
-        ECL_UNWIND_PROTECT_EXIT {                              \
-                mp_giveup_lock(__ecl_the_lock);               \
-        } ECL_UNWIND_PROTECT_END; }
-# define ECL_WITH_SPINLOCK_BEGIN(the_env,lock) {                \
-        const cl_env_ptr __ecl_the_env = (the_env);             \
-        cl_object *__ecl_the_lock = (lock);                     \
-        ecl_get_spinlock(__ecl_the_env, __ecl_the_lock);
-# define ECL_WITH_SPINLOCK_END                  \
-        ecl_giveup_spinlock(__ecl_the_lock); }
-#else
-# define ECL_WITH_GLOBAL_LOCK_BEGIN(the_env)
-# define ECL_WITH_GLOBAL_LOCK_END
-# define ECL_WITH_LOCK_BEGIN(the_env,lock)
-# define ECL_WITH_LOCK_END
-# define ECL_WITH_SPINLOCK_BEGIN(the_env,lock)
-# define ECL_WITH_SPINLOCK_END
-#endif /* ECL_THREADS */
-
-#ifdef ECL_RWLOCK
-# define ECL_WITH_GLOBAL_ENV_RDLOCK_BEGIN(the_env) {            \
-        const cl_env_ptr __ecl_pack_env = the_env;              \
-        ecl_disable_interrupts_env(__ecl_pack_env);             \
-        mp_get_rwlock_read_wait(cl_core.global_env_lock);
-# define ECL_WITH_GLOBAL_ENV_RDLOCK_END                   \
-        mp_giveup_rwlock_read(cl_core.global_env_lock);   \
-        ecl_enable_interrupts_env(__ecl_pack_env); }
-# define ECL_WITH_GLOBAL_ENV_WRLOCK_BEGIN(the_env) {            \
-        const cl_env_ptr __ecl_pack_env = the_env;              \
-        ecl_disable_interrupts_env(__ecl_pack_env);             \
-        mp_get_rwlock_write_wait(cl_core.global_env_lock);
-# define ECL_WITH_GLOBAL_ENV_WRLOCK_END                    \
-        mp_giveup_rwlock_write(cl_core.global_env_lock);   \
-        ecl_enable_interrupts_env(__ecl_pack_env); }
-#else
-# define ECL_WITH_GLOBAL_ENV_RDLOCK_BEGIN(the_env)
-# define ECL_WITH_GLOBAL_ENV_RDLOCK_END
-# define ECL_WITH_GLOBAL_ENV_WRLOCK_BEGIN(the_env)
-# define ECL_WITH_GLOBAL_ENV_WRLOCK_END
-#endif /* ECL_RWLOCK */
-
-#ifdef ECL_THREADS
-# define AO_REQUIRE_CAS
-# ifdef ECL_LIBATOMIC_OPS_H
-#  include <ecl/atomic_ops.h>
-# else
-#  include <atomic_ops.h>
-# endif
-#else
-# define AO_load(x) (x)
-# define AO_store(x,y) ((x)=(y))
+extern cl_object mp_get_rwlock_read_wait(cl_object lock);
+extern cl_object mp_get_rwlock_write_wait(cl_object lock);
 #endif
 
 /* read.d */
@@ -433,7 +350,6 @@ extern void cl_write_object(cl_object x, cl_object stream);
 #else
 #define RTABSIZE        ECL_CHAR_CODE_LIMIT     /*  read table size  */
 #endif
-extern cl_object si_make_backq_vector(cl_object dim, cl_object data, cl_object stream);
 
 /* package.d */
 
@@ -447,11 +363,11 @@ extern bool ecl_wild_string_p(cl_object item);
 typedef struct { cl_index start, end, length; } cl_index_pair;
 extern ECL_API cl_index_pair ecl_sequence_start_end(cl_object fun, cl_object s, cl_object start, cl_object end);
 
+#ifdef ECL_EXTERNALIZABLE
 /* serialize.d */
 
-extern cl_object si_serialize(cl_object root);
-extern cl_object si_deserialize(cl_object root);
 extern cl_object ecl_deserialize(uint8_t *data);
+#endif
 
 /* string.d */
 #define ecl_vector_start_end ecl_sequence_start_end
@@ -479,6 +395,13 @@ extern void ecl_cs_set_org(cl_env_ptr env);
 #ifdef ECL_THREADS
 extern ECL_API cl_object mp_suspend_loop();
 extern ECL_API cl_object mp_break_suspend_loop();
+
+# ifdef ECL_WINDOWS_THREADS
+#  define ecl_thread_exit() ExitThread(0);
+# else
+#  define ecl_thread_exit() pthread_exit(NULL);
+# endif  /* ECL_WINDOWS_THREADS */
+
 #endif
 
 /* time.d */
@@ -490,31 +413,127 @@ struct ecl_timeval {
 
 extern void ecl_get_internal_real_time(struct ecl_timeval *time);
 extern void ecl_get_internal_run_time(struct ecl_timeval *time);
-extern void ecl_musleep(double time, bool alertable);
+extern void ecl_musleep(double time);
 
 #define UTC_time_to_universal_time(x) ecl_plus(ecl_make_integer(x),cl_core.Jan1st1970UT)
 extern cl_fixnum ecl_runtime(void);
 
-/* threads/mutex.d */
+/* unixfsys.d */
+
+/* Filename encodings: on Unix we use ordinary chars encoded in a user
+ * specified format (usually utf8), while on Windows we use a wchar_t
+ * type.
+ *
+ * Naming conventions:
+ *  fstr: null-terminated raw C array with element type char or wchar_t
+ *  filename: Lisp base string or vector with element type byte16,
+ *            also null-terminated
+ */
+#if defined(ECL_MS_WINDOWS_HOST) && defined(ECL_UNICODE)
+#include <wchar.h>
+
+typedef wchar_t ecl_filename_char;
+#define ecl_fstrlen(x) wcslen(x)
+#define ecl_fstrcpy(x,y) wcscpy(x,y)
+#define ecl_fstrcat(x,y) wcscat(x,y)
+#define ecl_fstr(x) L ## x      /* wchar_t string constructor prefixed with L */
+
+cl_object ecl_make_simple_filename(const ecl_filename_char *x, cl_fixnum size);
+#define ecl_make_constant_filename(x,y) ecl_make_simple_filename(x,y)
+cl_object ecl_alloc_filename(cl_index len, cl_object adjustable);
+#define ecl_alloc_adjustable_filename(len) ecl_alloc_filename(len, ECL_T)
+#define ecl_alloc_simple_filename(len) ecl_alloc_filename(len, ECL_NIL)
+cl_object ecl_concatenate_filename(cl_object x, cl_object y);
+#define ecl_filename_self(x) ((ecl_filename_char*)((x)->vector.self.b16))
+
+#define ecl_chdir _wchdir
+#define ecl_stat _wstat64
+#define ecl_fstat _fstat64
+typedef struct __stat64 ecl_stat_struct;
+#define ecl_getcwd _wgetcwd
+#define ecl_access _waccess
+#define ecl_unlink _wunlink
+#define ecl_rename _wrename
+#define ecl_open _wopen
+#define ecl_fopen _wfopen
+#define ecl_fdopen _wfdopen
+#define ecl_rmdir _wrmdir
+#define ecl_mkdir _wmkdir
+#define ecl_chmod _wchmod
+#define ecl_getenv _wgetenv
+#define ecl_GetFileAttributes GetFileAttributesW
+#define ecl_MoveFile MoveFileW
+#define ecl_MoveFileEx MoveFileExW
+#define ecl_DeleteFile DeleteFileW
+#define ecl_FindFirstFile FindFirstFileW
+#define ecl_FindNextFile FindNextFileW
+#define ecl_WIN32_FIND_DATA WIN32_FIND_DATAW
+#define ecl_GetTempFileName GetTempFileNameW
+#define ecl_CopyFile CopyFileW
+#define ecl_LoadLibrary LoadLibraryW
+#define ecl_GetModuleFileName GetModuleFileNameW
+
+#else
+
+typedef char ecl_filename_char;
+#define ecl_fstrlen(x) strlen(x)
+#define ecl_fstrcpy(x,y) strcpy(x,y)
+#define ecl_fstrcat(x,y) strcat(x,y)
+#define ecl_fstr(x) x
+
+#define ecl_make_simple_filename(x,y) ecl_make_simple_base_string((char *)x,y)
+#define ecl_make_constant_filename(x,y) ecl_make_constant_base_string((char *)x,y)
+#define ecl_alloc_adjustable_filename(len) ecl_alloc_adjustable_base_string(len)
+#define ecl_alloc_simple_filename(len) ecl_alloc_simple_base_string(len)
+#define ecl_concatenate_filename(x,y) si_base_string_concatenate(2,x,y)
+#define ecl_filename_self(x) ((ecl_filename_char*)((x)->base_string.self))
+
+#define ecl_chdir chdir
+#define ecl_stat stat
+#define ecl_fstat fstat
+typedef struct stat ecl_stat_struct;
+#define ecl_getcwd getcwd
+#define ecl_access access
+#define ecl_unlink unlink
+#define ecl_rename rename
+#define ecl_open open
+#define ecl_fopen fopen
+#define ecl_fdopen fdopen
+#define ecl_rmdir rmdir
+#define ecl_mkdir mkdir
+#define ecl_chmod chmod
+#define ecl_getenv getenv
+#define ecl_GetFileAttributes GetFileAttributesA
+#define ecl_MoveFile MoveFileA
+#define ecl_MoveFileEx MoveFileExA
+#define ecl_DeleteFile DeleteFileA
+#define ecl_FindFirstFile FindFirstFileA
+#define ecl_FindNextFile FindNextFileA
+#define ecl_WIN32_FIND_DATA WIN32_FIND_DATAA
+#define ecl_GetTempFileName GetTempFileNameA
+#define ecl_CopyFile CopyFileA
+#define ecl_LoadLibrary LoadLibraryA
+#define ecl_GetModuleFileName GetModuleFileNameA
 
-#ifdef ECL_THREADS
-extern void ecl_process_yield(void);
-extern void print_lock(char *s, cl_object lock, ...);
-#define print_lock(...) ((void)0)
-extern void ecl_get_spinlock(cl_env_ptr env, cl_object *lock);
-extern void ecl_giveup_spinlock(cl_object *lock);
-extern cl_object ecl_wait_on(cl_env_ptr env, cl_object (*condition)(cl_env_ptr, cl_object), cl_object o);
-extern void ecl_wakeup_waiters(cl_env_ptr the_env, cl_object o, int flags);
-extern void ecl_wakeup_process(cl_object process);
-extern cl_object ecl_waiter_pop(cl_env_ptr the_env, cl_object q);
 #endif
 
-/* threads/rwlock.d */
+/*
+ * POSIX specifies that the "b" flag is ignored. This is good, because
+ * under MSDOS and Apple's OS we need to open text files in binary mode,
+ * so that we get both the carriage return and the linefeed characters.
+ * Otherwise, it would be complicated to implement file-position and
+ * seek operations.
+ */
+#define OPEN_R  ecl_fstr("rb")
+#define OPEN_W  ecl_fstr("wb")
+#define OPEN_RW ecl_fstr("r+b")
+#define OPEN_A  ecl_fstr("ab")
+#define OPEN_RA ecl_fstr("a+b")
+
+int ecl_backup_open(const ecl_filename_char *filename, int option, int mode);
+cl_object ecl_decode_filename(cl_object x, cl_object len);
+cl_object ecl_encode_filename(cl_object x, cl_object len);
 
-#ifdef ECL_RWLOCK
-extern cl_object mp_get_rwlock_read_wait(cl_object lock);
-extern cl_object mp_get_rwlock_write_wait(cl_object lock);
-#endif
 
 /* unixint.d */
 
@@ -525,8 +544,91 @@ extern cl_object mp_get_rwlock_write_wait(cl_object lock);
 
 extern void ecl_interrupt_process(cl_object process, cl_object function);
 
+/* disabling interrupts on the lisp side */
+
+#define ECL_WITHOUT_INTERRUPTS_BEGIN(the_env) do {                \
+        cl_env_ptr __the_env = (the_env);                         \
+        ecl_bds_bind(__the_env, ECL_INTERRUPTS_ENABLED, ECL_NIL);
+
+#define ECL_WITHOUT_INTERRUPTS_END                 \
+        ecl_bds_unwind1(__the_env);                \
+        ecl_check_pending_interrupts(__the_env); } while(0)
+
 /* unixsys.d */
-extern cl_object si_wait_for_all_processes _ECL_ARGS((cl_narg narg, ...));
+
+/* Some old BSD systems don't have WCONTINUED / WIFCONTINUED */
+
+#ifndef ECL_MS_WINDOWS_HOST
+# ifndef WCONTINUED
+#  define WCONTINUED 0
+# endif
+
+# ifndef WIFCONTINUED
+#  define WIFCONTINUED(x) 0
+# endif
+#endif /* ECL_MS_WINDOWS_HOST */
+
+/* global locks */
+
+#include <ecl/threads.h>
+
+#ifdef ECL_THREADS
+# define ECL_WITH_GLOBAL_LOCK_BEGIN(the_env)                    \
+        ECL_WITH_NATIVE_LOCK_BEGIN(the_env, &cl_core.global_lock)
+# define ECL_WITH_GLOBAL_LOCK_END               \
+        ECL_WITH_NATIVE_LOCK_END
+# define ECL_WITH_LOCK_BEGIN(the_env,lock) {             \
+        const cl_env_ptr __ecl_the_env = the_env;        \
+        const cl_object __ecl_the_lock = lock;           \
+        ecl_disable_interrupts_env(__ecl_the_env);       \
+        mp_get_lock_wait(__ecl_the_lock);                \
+        ECL_UNWIND_PROTECT_BEGIN(__ecl_the_env);         \
+        ecl_enable_interrupts_env(__ecl_the_env);
+# define ECL_WITH_LOCK_END                               \
+        ECL_UNWIND_PROTECT_THREAD_SAFE_EXIT {            \
+                mp_giveup_lock(__ecl_the_lock);          \
+        } ECL_UNWIND_PROTECT_THREAD_SAFE_END; }
+# define ECL_WITH_NATIVE_LOCK_BEGIN(the_env,lock) {      \
+        const cl_env_ptr __ecl_the_env = (the_env);      \
+        ecl_mutex_t* __ecl_the_lock = (lock);            \
+        ecl_disable_interrupts_env(__ecl_the_env);       \
+        ecl_mutex_lock(__ecl_the_lock);                  \
+        ECL_UNWIND_PROTECT_BEGIN(__ecl_the_env);         \
+        ecl_enable_interrupts_env(__ecl_the_env);
+# define ECL_WITH_NATIVE_LOCK_END                        \
+        ECL_UNWIND_PROTECT_THREAD_SAFE_EXIT {            \
+                ecl_mutex_unlock(__ecl_the_lock);        \
+        } ECL_UNWIND_PROTECT_THREAD_SAFE_END; }
+# define ECL_WITH_GLOBAL_ENV_RDLOCK_BEGIN(the_env) {      \
+        const cl_env_ptr __ecl_pack_env = the_env;        \
+        ecl_bds_bind(__ecl_pack_env, ECL_INTERRUPTS_ENABLED, ECL_NIL);  \
+        ecl_rwlock_lock_read(&cl_core.global_env_lock);
+# define ECL_WITH_GLOBAL_ENV_RDLOCK_END                   \
+        ecl_rwlock_unlock_read(&cl_core.global_env_lock); \
+        ecl_bds_unwind1(__ecl_pack_env);                  \
+        ecl_check_pending_interrupts(__ecl_pack_env); }
+# define ECL_WITH_GLOBAL_ENV_WRLOCK_BEGIN(the_env) {      \
+        const cl_env_ptr __ecl_pack_env = the_env;        \
+        ecl_bds_bind(__ecl_pack_env, ECL_INTERRUPTS_ENABLED, ECL_NIL);  \
+        ecl_rwlock_lock_write(&cl_core.global_env_lock);
+# define ECL_WITH_GLOBAL_ENV_WRLOCK_END                    \
+        ecl_rwlock_unlock_write(&cl_core.global_env_lock); \
+        ecl_bds_unwind1(__ecl_pack_env);                   \
+        ecl_check_pending_interrupts(__ecl_pack_env); }
+#else
+# define ECL_WITH_GLOBAL_LOCK_BEGIN(the_env)
+# define ECL_WITH_GLOBAL_LOCK_END
+# define ECL_WITH_LOCK_BEGIN(the_env,lock)
+# define ECL_WITH_LOCK_END
+# define ECL_WITH_NATIVE_LOCK_BEGIN(the_env,lock)
+# define ECL_WITH_NATIVE_LOCK_END
+# define ECL_WITH_GLOBAL_ENV_RDLOCK_BEGIN(the_env)
+# define ECL_WITH_GLOBAL_ENV_RDLOCK_END
+# define ECL_WITH_GLOBAL_ENV_WRLOCK_BEGIN(the_env)
+# define ECL_WITH_GLOBAL_ENV_WRLOCK_END
+#endif /* ECL_THREADS */
+
+#include <ecl/ecl_atomics.h>
 
 /*
  * Fake several ISO C99 mathematical functions if not available
@@ -539,6 +641,12 @@ extern cl_object si_wait_for_all_processes _ECL_ARGS((cl_narg narg, ...));
 # endif
 # define expf(x) exp((float)x)
 #endif
+#ifndef HAVE_POWF
+# ifdef powf
+#  undef powf
+# endif
+# define powf(x,y) pow((float)x,(float)y)
+#endif
 #ifndef HAVE_LOGF
 # ifdef logf
 #  undef logf
@@ -604,6 +712,25 @@ extern cl_object si_wait_for_all_processes _ECL_ARGS((cl_narg narg, ...));
 # define ldexpf(x,y) ldexp((float)x,y)
 #endif
 
+#ifndef HAVE_LOG1PF
+# ifdef log1pf
+#  undef log1pf
+# endif
+# define log1pf(x) logf(x+1.0f)
+#endif
+#ifndef HAVE_LOG1P
+# ifdef log1p
+#  undef log1p
+# endif
+# define log1p(x) log(x+1.0)
+#endif
+#ifndef HAVE_LOG1PL
+# ifdef log1pl
+#  undef log1pl
+# endif
+# define log1pl(x) logl(x+1.0l)
+#endif
+
 /*
  * Fake INFINITY and NAN defined in ISO C99 (portably)
  */
@@ -636,6 +763,19 @@ static union {
 # endif  /* _MSC_VER == 1600 */
 #endif /* ~NAN */
 
+#ifdef ECL_COMPLEX_FLOAT
+#include <complex.h>
+#ifndef CMPLXF
+# define CMPLXF(x, y) ((float complex)((float)(x) + I * (float)(y)))
+#endif
+#ifndef CMPLX
+# define CMPLX(x, y) ((double complex)((double)(x) + I * (double)(y)))
+#endif
+#ifndef CMPLXL
+# define CMPLXL(x, y) ((long double complex)((long double)(x) + I * (long double)(y)))
+#endif
+#endif
+
 #ifdef __cplusplus
 }
 #endif
